#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <mpi.h>

#include "func.h"

// programma dimostrativo deve essere migliorato se non altro
// nell'aspetto e nella gestione degli errori

int main (int argc, char ** argv) 
{
  int size, rank, rlen, n = -1, i;
  char myname[MPI_MAX_PROCESSOR_NAME]; 
  double * resvec, * vect, time, mtime, 
         ctime, commtime;

  MPI_Init (&argc, &argv);
  MPI_Comm_size (MPI_COMM_WORLD, &size);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Get_processor_name (myname, &rlen);

  fprintf (stdout, "I am proces %3d of %3d my name is %s (%s)\n", 
    rank, size, myname, argv[0]);
  fflush(stdout);

  if (argc == 2)
    n = atoi(argv[1]);

  if (n < 0)
    n = 100;

  if ((size * (int) (n/size)) != n) {
    fprintf (stderr, "Error\n");
    MPI_Finalize ();
    return EXIT_FAILURE;
  }

  if ( (vect = (double *) malloc (sizeof (double) * n)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    return EXIT_FAILURE;
  }

  if (rank == 0) {
    for (i=0; i<n; i++)
      vect[i] = 1.0e0 + (double)
        (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0)));

    if ( (resvec = (double *) malloc (sizeof (double) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
  }

  MPI_Barrier (MPI_COMM_WORLD);
  time = MPI_Wtime ();
  parallel (rank, size, n, vect, resvec, &ctime, 
      &commtime);
  MPI_Barrier (MPI_COMM_WORLD);
  time = MPI_Wtime () - time;

  MPI_Reduce (&time, &mtime, 1, MPI_DOUBLE, MPI_MAX, 0, 
      MPI_COMM_WORLD);
  MPI_Reduce (&commtime, &time, 1, MPI_DOUBLE, MPI_MAX, 0, 
      MPI_COMM_WORLD);
  commtime = time;
  MPI_Reduce (&ctime, &time, 1, MPI_DOUBLE, MPI_SUM, 0, 
      MPI_COMM_WORLD);
  ctime = time;

  if (rank == 0) {
    fprintf (stdout, "Parallel time: Tot %f s, Comp %f s, Comm %f s\n", 
        mtime, ctime, commtime);

    free (resvec);
  }

  free (vect);

  MPI_Finalize ();

  return EXIT_SUCCESS;
}
