#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#include "func.h"

void parallel (int rank, int size, int n, double * vect, 
    double * resvec, double * ctime, double * commtime) 
{
  int i, j, numofrow;
  double * rows, * result, time;

  numofrow = n/size;

  if ( (rows = (double *) malloc (sizeof (double) * n 
          * numofrow)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    exit (EXIT_FAILURE);
  }

  srand (rank);
  for (i=0; i<n*numofrow; i++) 
    rows[i] = 1.0e0 + (double) 
        (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0)));

  if ( (result = (double *) malloc (sizeof (double) * 
          numofrow)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    exit (EXIT_FAILURE);
  }

  time = MPI_Wtime ();
  MPI_Bcast (vect, n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  time = MPI_Wtime() - time;
  *commtime = time; 

  time = MPI_Wtime ();
  for (i=0; i<numofrow; i++) {
    result[i] = 0.0e0;
    for (j=0; j<n; j++)
      result[i] += vect[j] * rows[(n*i)+j];
  }
  time = MPI_Wtime() - time;

  *ctime = time;

  time = MPI_Wtime ();
  MPI_Gather (result, numofrow, MPI_DOUBLE, resvec, numofrow, MPI_DOUBLE, 
      0, MPI_COMM_WORLD );
  time = MPI_Wtime() - time;

  *commtime += time;

  free (rows);
  free (result);

  return;
}
