#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <mpi.h>

#include "func.h"

// programma dimostrativo deve essere migliorato se non altro
// nell'aspetto e nella gestione degli errori

int main (int argc, char ** argv) 
{
  int size, rank, rlen, n = -1, i, j;
  char myname[MPI_MAX_PROCESSOR_NAME]; 
  double ** d, * temp, * resvec, * vect, time, mtime, 
         ctime, commtime;

  MPI_Init (&argc, &argv);
  MPI_Comm_size (MPI_COMM_WORLD, &size);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Get_processor_name (myname, &rlen);

  fprintf (stdout, "I am proces %3d of %3d my name is %s (%s)\n", 
    rank, size, myname, argv[0]);
  fflush(stdout);

  if (argc == 2)
    n = atoi(argv[1]);

  if (n < 0)
    n = 100;

  if ( (vect = (double *) malloc (sizeof (double) * n)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    return EXIT_FAILURE;
  }

  if (rank == 0) {
    if ((d = (double **) malloc (sizeof (double *) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
    else {
      if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) 
          d[i] = &temp[i*n]; 
      }
    }

    srand (10);
    for (i=0; i<n; i++) 
      for (j=0; j<n; j++)
        d[i][j] = 1.0e0 + (double) 
          (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0)));

    for (i=0; i<n; i++)
      vect[i] = 1.0e0 + (double)
        (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0)));

    if ( (resvec = (double *) malloc (sizeof (double) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
  }

  MPI_Barrier (MPI_COMM_WORLD);
  time = MPI_Wtime ();
  parallel (rank, size, n, d, vect, resvec, &ctime, 
      &commtime);
  MPI_Barrier (MPI_COMM_WORLD);
  time = MPI_Wtime () - time;

  MPI_Reduce (&time, &mtime, 1, MPI_DOUBLE, MPI_MAX, 0, 
      MPI_COMM_WORLD);
  MPI_Reduce (&commtime, &time, 1, MPI_DOUBLE, MPI_MAX, 0, 
      MPI_COMM_WORLD);
  commtime = time;
  MPI_Reduce (&ctime, &time, 1, MPI_DOUBLE, MPI_SUM, 0, 
      MPI_COMM_WORLD);
  ctime = time;

  if (rank == 0)
    fprintf (stdout, "Parallel time: Tot %f s, Comp %f s, Comm %f s\n", 
        mtime, ctime, commtime);

  if (rank == 0) {
    double * c;

    if ( (c = (double *) malloc (sizeof (double) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
    }
    else {
      time = (double) clock();
      sequential (n, d, vect, c);
      time = (double) clock() - time;

      fprintf (stdout, "Sequential time: %f \n", time/CLOCKS_PER_SEC);

      for (i=0; i<n; i++) {
        //fprintf (stdout, "%f %f\n", c[i] , resvec[i]);
        if (fabs(c[i]-resvec[i]) > 1.0e-5) {
          fprintf (stdout, "Error at %d (%f)\n", i, c[i] - resvec[i]);
          break;
        }
      }

      free (c);
    }

    free (d);
    free (temp);
    free (resvec);
  }

  free (vect);

  MPI_Finalize ();

  return EXIT_SUCCESS;
}
