#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#include "func.h"

void parallel (int rank, int size, int n, double ** d, 
    double * vect, double * resvec, double * ctime,
    double * commtime) 
{
  int i, j, numofrow;
  double * rows, * result, time;

  numofrow = n/size;

  if ( (rows = (double *) malloc (sizeof (double) * n 
          * numofrow)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    exit (EXIT_FAILURE);
  }

  if ( (result = (double *) malloc (sizeof (double) * 
          numofrow)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    exit (EXIT_FAILURE);
  }

  time = MPI_Wtime ();
  MPI_Bcast (vect, n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  MPI_Scatter (d[0], n*numofrow, MPI_DOUBLE, rows, n*numofrow, MPI_DOUBLE, 
      0, MPI_COMM_WORLD); 
  time = MPI_Wtime() - time;
  *commtime = time; 

  time = MPI_Wtime ();
  for (i=0; i<numofrow; i++) {
    result[i] = 0.0e0;
    for (j=0; j<n; j++)
      result[i] += vect[j] * rows[(n*i)+j];
  }
  time = MPI_Wtime() - time;

  *ctime = time;

  time = MPI_Wtime ();
  MPI_Gather (result, numofrow, MPI_DOUBLE, resvec, numofrow, MPI_DOUBLE, 
      0, MPI_COMM_WORLD );
  time = MPI_Wtime() - time;

  *commtime += time;

  if (rank == 0) {
    if ((numofrow*size) != n) {
      time = MPI_Wtime ();
      for (i=(numofrow*size)-1; i<n; i++) {
        resvec[i] = 0.0e0;
        for (j=0; j<n; j++)
          resvec[i] += vect[j] * d[i][j]; 
      }
      time = MPI_Wtime() - time;
      *ctime += time;
    }
  }

  free (rows);
  free (result);

  return;
}
